import logging

from allensdk.internal.core.lims_pipeline_module import PipelineModule
from allensdk.internal.mouse_connectivity.interval_unionize.run_tissuecyte_unionize_cav import run 




def main():

    module = PipelineModule()

    input_data = module.input_data()
    output_data = run(input_data)
    module.write_output_data(output_data)


if __name__ == '__main__':
    main()
