import numpy as np

def are_two_lists_of_arrays_the_same(data1, data2):   
    '''returns False if to lists of arrays are different.
    otherwise the function returns True.
    '''
    
    if len(data1) != len(data2): 
        return False
    for a,b in zip(data1,data2):
        if np.any(a != b):
            return False
    
    return True
        
        